--[[
Auras.lua
Copyright (c) 2006 by Markus Inger, 2006 by Caeryn Dryad, 2007-2009 by Horacio Hoyos

This file is part of DHUD3.

    DHUD3 is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    DHUD3 is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with DHUD3.  If not, see <http://www.gnu.org/licenses/>.
]]

local DHUD3 = LibStub("AceAddon-3.0"):GetAddon("DHUD3");
local L = LibStub("AceLocale-3.0"):GetLocale("DHUD3");

local MODNAME = "Auras";
local Auras = DHUD3:NewModule(MODNAME, "AceEvent-3.0");

local math_ceil = math.ceil
local math_floor = math.floor
local string_match = string.match
local string_sub = string.sub
local pairs = pairs
local _G = _G
local auraUpdateInterval = 1
local auraPosition = "i";
local auraOrigin;
local defaultfont = "Fonts/FRIZQT__.TTF"

--Local Defaults
local db
local defaults = {
	profile = {
        side = "l";
		scale = 1;
        onlyMine = false;
        filter = false;
        filterTime = 30;
		buffsTips = true;
		fontSize = 10;
        weapon = true;
        weaponForce = true; -- Alwyas save space for weapon buffs
		weaponOrder = "last";
        expMin = 0;
        expSec = 59;
        colors = {
            full = { 1, 1, 1 },
            med = { 255/255, 125/255, 125/255 },
            low = { 255/255, 65/255, 0 }      
        },
    }
}

local OptGetter, OptSetter, ColorGetter, ColorSetter, WeaponHidden, WeaponBuff
do
	local mod = Auras
	function OptGetter(info)
		local key = info[#info];
		return db[key];
	end

	function OptSetter(info, value)
		local key = info[#info];
		db[key] = value;
		mod:Refresh();
	end
    
    function ColorGetter (info)
        local key1, key2 = info[#info-1], info[#info];
		return db[key1][key2][1], db[key1][key2][2], db[key1][key2][3];
	end
    
    function ColorSetter (info, r, g, b) 
        local key1, key2, key3 = info[#info-2], info[#info-1], info[#info];
        db[key1][key2][key3][1] = r;
        db[key1][key2][key3][2] = g;
        db[key1][key2][key3][3] = b;
        mod:Refresh();
    end
    
	function WeaponHidden()
		return not db.weapon;
	end
	
    WeaponBuff = {["first"] = L["First"], ["last"] = L["Last"], ["lower"] = L["Lower"], ["higer"] = L["Higher"]}
end

--Local Options
local options
local function getOptions()
	if not options then
        --Options table
    	options = {
    		type = "group",
			name = L["Auras"],
			arg = MODNAME,
			get = OptGetter,
			set = OptSetter,
    		args = {
                enabled = {
                    type = "toggle",
                    order = 1,
                    name = L["Enable Auras Module"],
                    get = function() return DHUD3:GetModuleEnabled(MODNAME) end,
                    set = function(info, value) DHUD3:SetModuleEnabled(MODNAME, value) end,
                },
				general = {
					type = 'group',
					name = L["General"],
					order = 2,
					inline = true,
					args = {
						desc = {
							order = 0,
							type = "description",
							name = L["Configure general functionalily"],
						},
						side = {
							type = 'select',
							order = 1,
							name = L["Auras Side"],
							values = {["r"] = L["Right"], ["l"] = L["Left"]},
						},
						buffsTips = {
							type = 'toggle',
							order = 2,
							name = L["Show Aura tips"],
						},
						fontSize = {
							type = 'range',
							order = 3,
							name = L["Font Size"],
							min = 6,
							max = 25,
							step = 1,
						},
						expMin = {
							type = 'range',
							order = 4,
							name = L["Expiration Minutes"],
							min = 0,
							max = 59,
							step = 1,
						},
				        expSec = {
							type = 'range',
							order = 5,
							name = L["Expiration Seconds"],
							min = 0,
							max = 59,
							step = 1,
						},
						colors = { 
							type = 'group',
							name = L["Border Color"],
							order = 7,
							guiInline  = true,
							get = ColorGetter,
		                    set = ColorSetter,
							args = {
								full = {
									type = 'color',
									name = L["Full"],
									order = 1,
								},
								med = {
									type = 'color',
									name = L["Med"],
									order = 2,
								},
								low = {
									type = 'color',
									name = L["Low"],
									order = 3,
								},
							},
						},
						scale = {
							type = "range",
							order = 6,
							name = L["Set Auras's scale"],
							min = 0.5,
							max = 1,
							step = 0.05,
						},
					},
				},
				filters = {
					type = 'group',
					order = 3,
					name = L["Filters"],
					inline = true,
					args =  {
						desc = {
							order = 0,
							type = "description",
							name = L["Aura filter options"],
						},
						onlyMine = {
							type = 'toggle',
							order = 1,
							name = L["Only Mine"],
							desc = L["Only show own auras"],
						},
				        filter = {
							type = 'toggle',
							order = 2,
							name = L["Filter Auras"],
							desc = L["Filter long duration Auras"],
						},
						filterTime = {
							type = 'range',
							order = 3,
							name = L["Filter Auras duration"],
							desc = L["Filter Auras with a total duration higher than this value (minutes)"],
							min = 10,
							max = 59,
							step = 1,
							disabled = function() return not db.filter; end,
						},
					},
				},
				weapons = {
					type = 'group',
					order = 4,
					name = L["Weapons"],
					inline = true,
					args = {
						desc = {
							order = 0,
							type = "description",
							name = L["Weapon enchant display"],
						},
				        weapon = {
							type = 'toggle',
							order = 1,
							name = L["Weapon Enchants"],
							desc = L["Show/Hide Auras from weapon enchants"],
						},
				        weaponOrder = {
							type = 'select',
							order = 2,
							name = L["Weapon Auras order"],
							desc = L["Choose weapon Auras display order"],
							values = WeaponBuff,
							disabled = WeaponHidden,
						},
					},
				},
            },
        }
        
    end
    
    return options;
end

local UpdatePlayerAuras, HidePlayerBuffs, OnUpdateAura
do
	function UpdatePlayerAuras()
		--DHUD3:Debug("updatePlayerAuras");
		-- Auras can be buffs or Weapon Enchants
	    local weaponInsert = {};
		weaponInsert.mainHand = false;
		weaponInsert.offHand = false;
	    local timeToEvent = 0;
		local showTime = (db.expMin * 60) + db.expSec;
		local showFilter = db.filterTime * 60;
		local timeLeft, hasWeapon, id, name,texture, duration, expirationTime, caster;
		local auraInit = 1;
		local hasMainHandEnchant, mainHandExpiration, _, hasOffHandEnchant, offHandExpiration = GetWeaponEnchantInfo();
		local insertWeapon = true;
		if db.weapon then
			-- Weapons
			local mainHand, offHand;
			if ( db.weaponOrder == "first" ) then
				mainHand = 1;
				offHand = hasMainHandEnchant and 2 or 1;
			elseif ( db.weaponOrder == "last") then
				insertWeapon = false;
			elseif ( db.weaponOrder == "lower" ) then
				mainHand = 1;
				offHand = hasMainHandEnchant and 9 or 1;
			elseif ( db.weaponOrder == "higer" ) then
				mainHand = 8;
				offHand = hasMainHandEnchant and 16 or 8;
			end
			if ( hasMainHandEnchant ) then
				local texture = GetInventoryItemTexture("player", 16);
				weaponInsert.mainHand = { mainHand, mainHandExpiration/1000, texture, 16 };
			end
			if ( hasOffHandEnchant ) then
				local texture = GetInventoryItemTexture("player", 17);
				weaponInsert.offHand = { offHand, offHandExpiration/1000, texture, 17 };
			end
		end
		
		local moreAuras = true;
		for j = 1, 16 do
			local show = false;
			hasWeapon = false;
			if ( moreAuras ) then
				for i = auraInit, 40 do
					name, _, texture, _, _, duration, expirationTime, caster = UnitBuff("player", i);
					if ( name ) then
						if ( not ( ( caster ~= "player" ) and db.onlyMine ) ) then
							if not ( db.filter and ( duration > showFilter ) ) then
								timeLeft = expirationTime - GetTime();
								id = i;
								auraInit = i + 1;
								show = true;
								break;
							end
						end
					end
				end	
				if ( not show ) and ( not insertWeapon ) then
					
					moreAuras = false;
					if ( hasMainHandEnchant ) then
						local texture = GetInventoryItemTexture("player", 16);
						weaponInsert.mainHand = { j, mainHandExpiration/1000, texture, 16 };
					end
					if ( hasOffHandEnchant ) then
						local texture = GetInventoryItemTexture("player", 17);
						weaponInsert.offHand = { hasMainHandEnchant and j + 1 or j, offHandExpiration/1000, texture, 17 };
					end
				end
			end
			if ( weaponInsert.mainHand ) then
				if ( weaponInsert.mainHand[1] == j ) then
					timeLeft = weaponInsert.mainHand[2];
					texture = weaponInsert.mainHand[3];
					id = weaponInsert.mainHand[4];
					if ( show ) then auraInit = auraInit - 1; end
					show = true;
					hasWeapon = true;
				end
			elseif ( weaponInsert.offHand ) then
				if ( weaponInsert.offHand[1] == j ) then
					timeLeft = weaponInsert.offHand[2];
					texture = weaponInsert.offHand[3];
					id = weaponInsert.offHand[4];
					if ( show ) then auraInit = auraInit - 1; end
					show = true;
					hasWeapon = true;
				end
			end
			if ( show ) then
				if ( timeLeft > 0 and math_floor(timeLeft) <= showTime ) then
					local r, g, b, hex = DHUD3:Colorize(db.colors, timeLeft / 20);
					local button = _G["DHUD3_"..db.side.."buff"..j];
					button.id = id;
					button.hasweapon = hasWeapon;
					button.auraTimer = 0;
					button.left = timeLeft;
					button:SetNormalTexture(texture);
					local buffBorder = _G["DHUD3_"..db.side.."buff"..j.."_Border"];
					buffBorder:SetVertexColor(r, g, b);
					buffBorder:Show();
					local buffText = _G["DHUD3_"..db.side.."buff"..j.."_Text"];
					buffText:SetText("|cff"..hex..DHUD3:FormatTime(button.left));
					local w = buffText:GetStringWidth() + 10;
					buffText:SetWidth(w);
					button:SetScript("OnUpdate", OnUpdateAura);
					button:EnableMouse(true);
					button:Show();
				elseif timeLeft > 0 then
					local newtimeToEvent = math_ceil(timeLeft) - showTime
					if ( ( newtimeToEvent < timeToEvent ) or timeToEvent == 0 ) then
						timeToEvent = newtimeToEvent;
					end
					local button = _G["DHUD3_"..db.side.."buff"..j]
			        button.id = 0
			        button:SetScript("OnUpdate", nil)
			        button:EnableMouse(false)
			        button:Hide()
				end
			else
				local button = _G["DHUD3_"..db.side.."buff"..j]
		        button.id = 0
		        button:SetScript("OnUpdate", nil)
		        button:EnableMouse(false)
		        button:Hide()
			end
		end
		
		
	    if timeToEvent > 1 then
	        -- Trigger an event for the buff with closest time to playerBuffTimeFilter
	        DHUD3_mainFrame.buffTimer = timeToEvent
			DHUD3_mainFrame.crono = 0
	        DHUD3_mainFrame.buffTimerOn = true
	        DHUD3_mainFrame:SetScript("OnUpdate", DHUD3.OnUpdate)
	    else
	        --DHUD3:Debug("No Timer")
	        DHUD3_mainFrame.buffTimerOn = false
	        DHUD3_mainFrame:SetScript("OnUpdate", nil)
	    end
		
	end

	function HidePlayerBuffs()
		for i = 1, 16 do
			 _G["DHUD3_lbuff"..i]:Hide()
			 _G["DHUD3_rbuff"..i]:Hide()
		end
	end

	function OnUpdateAura(event, elapsed)
	    this.auraTimer = this.auraTimer + elapsed
	    if this:GetEffectiveAlpha() == 0 then
	        this:EnableMouse(false)
	    else            
	        this:EnableMouse(true)
	        if this.auraTimer >= auraUpdateInterval then
				local timeLeft = 0;
	            if this.hasweapon then
	                local _, mainHandExpiration, _, _, offHandExpiration, _ = GetWeaponEnchantInfo()
	                if this.id == 16 and mainHandExpiration then 
	                    timeLeft = mainHandExpiration/1000
	                elseif this.id == 17 and offHandExpiration then
	                    timeLeft = offHandExpiration/1000
	                end
				else
	                this.left = this.left - this.auraTimer
					timeLeft = this.left
	            end
	            if timeLeft <= 0 then
					this.id = 0
	                this.hasbuff = false
	                this.hasweapon = false
	                this:EnableMouse(false)
	                this:SetScript("OnUpdate", nil)
	                return this:Hide()
	            end
				this.auraTimer = 0
	            local r, g, b, hex = DHUD3:Colorize(db.colors, this.left / 20)
	            local buffBorder = _G[this:GetName().."_Border"]
	            buffBorder:SetVertexColor(r, g, b)
	            local buffText = _G[this:GetName().."_Text"]
	            buffText:SetText("|cff"..hex..DHUD3:FormatTime(this.left))
	            local w = buffText:GetStringWidth() + 10
	            buffText:SetWidth(w)
				if ( GameTooltip:IsOwned(this) ) then
					if ( not this.hasWeapon ) then
						GameTooltip:SetUnitBuff("player", this.id);
					else
						GameTooltip:SetInventoryItem("player", this.id);
					end
				end						
	        end
	    end
	end

end

function Auras:OnInitialize()

    self.db = DHUD3.db:RegisterNamespace(MODNAME, defaults);
	db = self.db.profile;
	self:SetEnabledState(DHUD3:GetModuleEnabled(MODNAME));
	DHUD3:RegisterModuleOptions(MODNAME, getOptions, L["Auras"]);

	local playBuff = DHUD3.layout.playbuff
    -- Player Buff
	for k = 1, 16 do
		-- Left
        ref = CreateFrame("Button", "DHUD3_lbuff"..k, _G["DHUD3_leftAura"]);
        ref:ClearAllPoints();
        ref:SetPoint("BOTTOM", _G["DHUD3_leftAura"], "BOTTOM", 0, 0);
        ref:SetHeight(26);
        ref:SetWidth(26);
        font = ref:CreateFontString("DHUD3_lbuff"..k.."_Text", "OVERLAY");
        font:SetFontObject(GameFontHighlightSmall);
        font:ClearAllPoints();
        font:SetPoint("CENTER", ref, "CENTER", 0,0);
        font:SetText("");
        font:SetJustifyH("CENTER");
        font:SetJustifyV("CENTER");
        font:SetWidth(26*2);
        bgt = ref:CreateTexture("DHUD3_lbuff"..k.."_Border", "OVERLAY");
        bgt:SetTexture("Interface\\AddOns\\DHUD3\\textures\\serenity0");
        bgt:SetVertexColor(1.0, 1.0, 1.0);
        bgt:SetPoint("CENTER", ref, "CENTER", 0, 0);
        bgt:SetHeight(26*1.7);
        bgt:SetWidth(26*1.7);
        bgt:SetTexCoord(0, 1, 0, 1);
        bgt:SetBlendMode("BLEND");
        ref:SetNormalTexture( "Interface\\AddOns\\DHUD3\\icon" );
		ref:Hide();
        
		-- Right
		ref = CreateFrame("Button", "DHUD3_rbuff"..k, _G["DHUD3_rightAura"]);
        ref:ClearAllPoints();
        ref:SetPoint("BOTTOM", _G["DHUD3_rightAura"], "BOTTOM", 0, 0);
        ref:SetHeight(26);
        ref:SetWidth(26);
        font = ref:CreateFontString("DHUD3_rbuff"..k.."_Text", "OVERLAY");
        font:SetFontObject(GameFontHighlightSmall);
        font:ClearAllPoints();
        font:SetPoint("CENTER", ref, "CENTER", 0,0);
        font:SetText("");
        font:SetJustifyH("CENTER");
        font:SetJustifyV("CENTER");
        font:SetWidth(26*2);
        bgt = ref:CreateTexture("DHUD3_rbuff"..k.."_Border", "OVERLAY");
        bgt:SetTexture("Interface\\AddOns\\DHUD3\\textures\\serenity0");
        bgt:SetVertexColor(1.0, 1.0, 1.0);
        bgt:SetPoint("CENTER", ref, "CENTER", 0, 0);
        bgt:SetHeight(26*1.7);
        bgt:SetWidth(26*1.7);
        bgt:SetTexCoord(0, 1, 0, 1);
        bgt:SetBlendMode("BLEND");
        ref:SetNormalTexture( "Interface\\AddOns\\DHUD3\\icon" );
        ref:Hide();
    end
    
end

function Auras:OnEnable()
    
    self:RegisterEvent("UNIT_AURA");
	self:Refresh();
	if ( UnitExists("target") ) then
		self:DHUD3_TARGETBARS_SHOW();
	end
	if DHUD3:GetModuleEnabled("Player") then
		Auras:RegisterMessage("DHUD3_BAR_SHOW");
		Auras:RegisterMessage("DHUD3_BAR_HIDE");
	end
	
end

function Auras:OnDisable()

    self:UnregisterAllEvents();
	for i = 1, 16 do
		local button = _G["DHUD3_"..db.side.."buff"..i];
		button:SetScript("OnUpdate", nil);
		button:Hide();
	end
end

function Auras:Refresh()

	db = self.db.profile;
	if not self:IsEnabled() then return; end

	DHUD3_leftAura:SetScale(db.scale);
	DHUD3_rightAura:SetScale(db.scale);
	
	local hide, parent, xDev;
	if ( db.side == "l" ) then
		parent = _G["DHUD3_leftAura"];
		hide = "r";
		xDev = - ( 151 / db.scale) - 13 * db.scale;
	else
		parent = _G["DHUD3_rightAura"];
		hide = "l";
		xDev = ( 151 / db.scale) + 13 * db.scale;
	end
	local x, y;
	local xCol = 0;
	for i = 1, 16 do
		-- Show
		local name = "DHUD3_"..db.side.."buff"..i;
		local button = _G[name];
		button:ClearAllPoints();
		step = ( 256 * db.scale)/8;
		if (i < 9 ) then
			y = (step * (i - 0.5));
			yPos = (32 * ( i - 0.5)) - 15;
		else
			y = (step * (i - 8.5));
			xCol = (db.side == "l") and -35 or 35;
			yPos = (32 * ( i - 8.5)) - 15;
		end
		if ( db.side == "l" ) then
			x = (-17.899 - 0.351*y + 0.00137*y^2)/db.scale;
		else
			x = (17.899 + 0.351*y - 0.00137*y^2)/db.scale;
		end
        button:SetPoint("BOTTOM", parent, "BOTTOM", x + xDev + xCol, yPos);
		button:SetScript("OnUpdate", nil);
		button:Hide();
		font = _G[name.."_Text"]
        font:SetFont(defaultfont, db.fontSize);
        local w = font:GetStringWidth() * 1.1;
        font:SetWidth(w);
		button:SetScript("OnEnter", function(this) 
                if not (this:IsVisible() and db.buffsTips) or (this:GetEffectiveAlpha() == 0) then
                    return;
                end
                GameTooltip:SetOwner(this, "ANCHOR_BOTTOMRIGHT");
                GameTooltip:SetUnitAura("player", this.id, "HELPFUL");
            end )
        button:SetScript("OnLeave", function() 
                GameTooltip:Hide();
            end )
		
		--Hide
		local name = "DHUD3_"..hide.."buff"..i;
		button = _G[name];
		button:SetScript("OnUpdate", nil);
		button:SetScript("OnEnter", nil);
		button:Hide();
        font = _G[name.."_Text"];
        font:SetFont(defaultfont, db.fontSize * db.scale)
        w = font:GetStringWidth() * 1.1
        font:SetWidth(w);
	end		

	UpdatePlayerAuras();
end

function Auras:UNIT_AURA(event, who)
	----self:Debug("UNIT_AURA", who)
	if who == "player" then
		UpdatePlayerAuras();
    end
end

function Auras:DHUD3_BAR_SHOW(event, bar)
	
	--DHUD3:Debug("DHUD3_BAR_SHOW", bar);
	local side = string_sub(bar, 1, 1);
	local position = string_sub(bar, 2);
	if ( side == db.side and position ~= auraPosition ) then
		-- The bar popped in the same side the auras are and the auras need to be moved
		local xDis = 0;
		local move = false;
		if ( auraPosition == "i" ) then
			if ( position == "m" ) then
				xDis = 15;
				move = true;
				auraOrigin = auraPosition;
				auraPosition = position;
			elseif ( position == "o" ) then
				xDis = 30;
				move = true;
				auraOrigin = auraPosition;
				auraPosition = position;
			end
		elseif ( auraPosition == "m" ) then
			if ( position == "o" ) then
				xDis = 15;
				move = true;
				auraOrigin = auraPosition;
				auraPosition = position;
			end
		end
		if ( move ) then
			for i = 1, 16 do
				local name = "DHUD3_"..db.side.."buff"..i;
				local button = _G[name];
				local point, relativeTo, relativePoint, xOfs, yOfs = button:GetPoint(1);
				button:ClearAllPoints();
		        button:SetPoint(point, relativeTo, relativePoint, xOfs + xDis, yOfs)
			end
		end
	end
end

function Auras:DHUD3_BAR_HIDE(event, bar)
	
	--DHUD3:Debug("DHUD3_BAR_HIDE", bar);
	local side = string_sub(bar, 1, 1);
	local position = string_sub(bar, 2);
	if ( side == db.side and position == auraPosition ) then
		-- The bar popped in the same side the auras are and the auras need to be moved
		local xDis = 0;
		local move = false;
		if ( auraPosition == "m" ) then
				xDis = -15;
				move = true;
				auraPosition = auraOrigin;
		elseif ( auraPosition == "o" ) then
			if ( auraOrigin == "i" ) then
				xDis = -30;
				move = true;
				auraPosition = "i";
			elseif ( auraOrigin == "m" ) then
				xDis = -15;
				move = true;
				auraPosition = "m";
			end
		end
		if ( move ) then
			for i = 1, 16 do
				local name = "DHUD3_"..db.side.."buff"..i;
				local button = _G[name];
				local point, relativeTo, relativePoint, xOfs, yOfs = button:GetPoint(1);
				button:ClearAllPoints();
		        button:SetPoint(point, relativeTo, relativePoint, xOfs - 14, yOfs)
			end
		end
	elseif ( side == db.side and position ~= auraPosition ) then
		-- The bar popped but the auras need not to be moved, verify if the origin needs to change
		if ( auraPosition == "o" and position == "m") then
			auraOrigin = "i";
		end		
	end
end

